/*
 * Copyright (C) 2007-2018 University of Stuttgart, IKR
 * 
 * This file is part of IKR SimLib, Java Edition, as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License (LGPL) as published by the
 * Free Software Foundation, in version 2.1 as it comes in the file "COPYING" in
 * directory doc of IKR SimLib.
 * 
 * IKR SimLib is distributed in the hope that it will be useful, but without any
 * warranty of any kind. See the LGPL for details or contact the IKR if you need
 * additional information (http://www.ikr.uni-stuttgart.de).
 * 
 * Alternatively you may license IKR SimLib under a different (non-free)
 * license. Please contact the IKR for further details.
 */

package ikr.simlib.example.ackNackModel;

import ikr.simlib.SimLibFormatter;
import ikr.simlib.control.SimTreeApplication;
import ikr.simlib.control.simulationcontrol.ControlTimer;
import ikr.simlib.distributions.continuous.ContinuousDistribution;
import ikr.simlib.distributions.discrete.DiscreteConstantDistribution;
import ikr.simlib.entities.generators.Generator;
import ikr.simlib.entities.generators.StdGenerator;
import ikr.simlib.entities.sinks.Sink;
import ikr.simlib.events.time.Duration;
import ikr.simlib.example.ackNackModel.components.Channel;
import ikr.simlib.example.ackNackModel.components.Receiver;
import ikr.simlib.example.ackNackModel.components.Sender;
import ikr.simlib.meters.time.TimeMeter;
import ikr.simlib.model.Model;
import ikr.simlib.model.SimNode;
import ikr.simlib.parameters.Parameters;
import ikr.simlib.random.RandomNumberGenerator;
import ikr.simlib.statistics.Statistics;
import ikr.simlib.statistics.UnweightedSampleCollector;
import ikr.simlib.statistics.distribution.DistributionStatistic;
import ikr.simlib.statistics.distribution.trace.TracingDistributionStatistic;
import ikr.simlib.statistics.sample.SampleStatistic;
import ikr.simlib.statistics.sample.StdSampleStatistic;

public class AckNackModel extends Model {

   public AckNackModel(Parameters pars) {
      super("AckNackModel");

      final ContinuousDistribution iatDistribution = pars.get(simNode, "IAT").asInstance(ContinuousDistribution.class, this.simNode.createChildNode("IAT"), pars, RandomNumberGenerator.getSystemRNG());
      final Generator generator = new StdGenerator(iatDistribution, new DiscreteConstantDistribution(1), this.simNode.createChildNode("Generator"), null, true);
      final Sender sender = new Sender(this.simNode.createChildNode("Sender"));
      final Receiver receiver = new Receiver(this.simNode.createChildNode("Receiver"), pars);
      final Sink sink = new Sink(this.simNode.createChildNode("Sink"));

      final Channel forwardChannel = new Channel(this.simNode.createChildNode("ForwardChannel"), pars);
      final Channel backwardChannel = new Channel(this.simNode.createChildNode("BackwardChannel"), pars);

      generator.getOutput().connect(sender.getInput());
      sender.getOutput().connect(forwardChannel.getInput());
      forwardChannel.getOutput().connect(receiver.getInput());
      receiver.getOutput().connect(sink.getInput());
      receiver.getFeedbackOutput().connect(backwardChannel.getInput());
      backwardChannel.getOutput().connect(sender.getFeedbackInput());

      final SimNode tranmissionTimeStatNode = this.simNode.createChildNode("TranmsissionTime");
      final SampleStatistic transmissionTimeSample = new StdSampleStatistic(tranmissionTimeStatNode.createChildNode("Statistic"));
      final DistributionStatistic transmissionTimeDistribution = new TracingDistributionStatistic(tranmissionTimeStatNode.createChildNode("Distribution"), 100);
      final UnweightedSampleCollector<Double> transmisisonTimeStat = Statistics.combine(transmissionTimeSample, transmissionTimeDistribution);
      final TimeMeter timeMeter = TimeMeter.createWithDoubleStatistic(transmisisonTimeStat);
      timeMeter.attachFromPort(generator.getOutput());
      timeMeter.attachToPort(sink.getInput());

      @SuppressWarnings("unused")
      final ControlTimer controlTimer = new ControlTimer(Duration.fromSeconds(100), Duration.fromSeconds(100));
   }

   public static void main(String[] args) {
      SimLibFormatter.setupAsFormatterForANewRootConsoleHandler();
      SimTreeApplication app = new SimTreeApplication("AckNackModelApplication", args) {
         @Override
         protected Model createModel(Parameters pars) {
            return new AckNackModel(pars);
         }
      };
      app.run();
   }

}
